﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-07-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-08-2014
// ***********************************************************************
// <copyright file="Animation.cs" >
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Autodesk.Navisworks.Api;
using ComBridge = Autodesk.Navisworks.Api.ComApi.ComApiBridge;
using COMApi = Autodesk.Navisworks.Api.Interop.ComApi;
using Autodesk.Navisworks.Api.Interop.ComApi;
using System.Diagnostics;

/// <summary>
/// The Models namespace includes all the models and interfaces in the MVVM arcitechture. 
/// </summary>
namespace DynamicRelaxationClashVerification.Models
{
    /// <summary>
    /// The Animation class is not currently implemented in the plugin application. There is
    /// however presented some method to show how the Navisworks API can be used to create 
    /// some animations. As the Navisworks .NET API 2014 not currently have functionality 
    /// to generate animations, the methods only show how to alter the camera, create and 
    /// save viewpoints, etc.
    /// </summary>
    public class Animation
    {

        /// <summary>
        /// Gets the view point list.
        /// </summary>
        /// <value>The view point list.</value>
        public List<Viewpoint> ViewPointList { get; private set; }
        /// <summary>
        /// The current NavisWorks document;
        /// </summary>
        private Document oDoc = Autodesk.Navisworks.Api.Application.ActiveDocument;

        /// <summary>
        /// Initializes a new instance of the <see cref="Animation"/> class.
        /// </summary>
        public Animation()
        {
            ViewPointList = new List<Viewpoint>();
        }
        /// <summary>
        /// Adds the view point to view point list.
        /// </summary>
        public void AddViewPointToAnimation()
        {
            // get current viewpoint
            Viewpoint oCurVP = oDoc.CurrentViewpoint.CreateCopy();
            ViewPointList.Add(oCurVP);
            
        }
        /// <summary>
        /// Creates all the saved viewpoints from list, as
        /// view points in Navisworks. The viewpoints can than be used
        /// in Navisworks to create a animation manually.
        /// </summary>
        public void CreateAnimation()
        {
            oDoc.SavedViewpoints.Clear();

            foreach (Viewpoint point in ViewPointList)
            {
                SavedViewpoint sVpoint = new SavedViewpoint(point);
                oDoc.SavedViewpoints.AddCopy(sVpoint);

            }
        }
        /// <summary>
        /// Example: Exports the animation to avi.
        /// Source: Autodesk Navisworks forum.
        /// </summary>
        public void ExportAnimationToAVI()
        {
            //Set the IOPlugin options
            String ioPluginName;

            ioPluginName = "lcodpanim";

            COMApi.InwOpState10 state = ComBridge.State;
            COMApi.InwOaPropertyVec options = state.GetIOPluginOptions(ioPluginName);
            //ListPropertyVec(ref options);

            foreach (COMApi.InwOaProperty opt in options.Properties())
            {
                if (opt.name == "export.image.anit-alias.level")
                    opt.value = 4;
                else if (opt.name == "export.anim.fps")
                    opt.value = 30;
                else if (opt.name == "export.image.height")
                    opt.value = 1280;
                else if (opt.name == "export.image.width")
                    opt.value = 720;

                //Timeliner Test
                else if (opt.name == "source")
                    opt.value = "LcTlAnimSequencyFactory";
                else if (opt.name == "renderer")
                    opt.value = "lcodpopengl";
            }


            //Export the animation
            String exportFilePath = "C:/timeliner_test.avi";
            state.DriveIOPlugin(ioPluginName, exportFilePath, options);
        }

        /*
         * The following methods are collected from 
         * articles about Viewpoints in Navisworks .NET
         * API 2013 by Xiaodong Liang, developer at 
         * Autodesk. They are only ment to be examples
         * to the developer in further work with the plugin
         * application.
         */

        /// <summary>
        /// Dumps the view point.
        /// by Xiaodong Liang, developer at Autodesk.
        /// </summary>
        private void dumpViewPoint()
        {
            // get current viewpoint
            Viewpoint oCurVP = oDoc.CurrentViewpoint;
 
            // common properties
            Debug.Print("Far Plane Distance: " +
                    oCurVP.FarPlaneDistance);
            Debug.Print("Near Plane Distance: " +
                    oCurVP.NearPlaneDistance);
            Debug.Print("Projection: " +
                    oCurVP.Projection);
            Debug.Print("Aspect Ratio: " +
                    oCurVP.AspectRatio);
 
 
            if (oCurVP.HasLinearSpeed)
                Debug.Print("Linear Speed: " +
                    oCurVP.LinearSpeed);
            else
                Debug.Print("Linear Speed: <None>");
            if (oCurVP.HasAngularSpeed)
                Debug.Print("Angular Speed: " +
                    oCurVP.AngularSpeed);
            else
                Debug.Print("Angular Speed: <None>");
            if (oCurVP.HasLighting)
                Debug.Print("Lighting: " +
                    oCurVP.Lighting);
            else
                Debug.Print("Lighting: <None>");
 
            if (oCurVP.HasRenderStyle)
                Debug.Print("RenderStyle: " +
                    oCurVP.RenderStyle);
            else
                Debug.Print("RenderStyle:<None> " +
                    oCurVP.RenderStyle);
 
            if (oCurVP.HasTool)
                Debug.Print("Tool: " +
                    oCurVP.Tool);
            else
                Debug.Print("Tool:<None> " +
                    oCurVP.Tool);
 
 
            // camera properties
            double oFocal =
                oCurVP.FocalDistance;
            Debug.Print("Focal Distance: " +
                oFocal);
            // Rotation
            Rotation3D oRot = oCurVP.Rotation;
            Debug.Print("Quaternion: <A (x),B(y),C(z),D(w)> = <{0},{1},{2},{3}>",
                          oRot.A,
                          oRot.B,
                          oRot.C,
                          oRot.D);
 
            // calculate view direction
            Rotation3D oNegtiveZ =
                new Rotation3D(0, 0, -1, 0);
            Rotation3D otempRot =
                MultiplyRotation3D(oNegtiveZ, oRot.Invert());
            Rotation3D oViewDirRot =
                MultiplyRotation3D(oRot, otempRot);
            // get view direction
            Vector3D oViewDir =
                new Vector3D(oViewDirRot.A,
                            oViewDirRot.B,
                            oViewDirRot.C);
            oViewDir.Normalize();
            Debug.Print("View Direction:<{0},{1},{2}> ",
                        oViewDir.X,
                        oViewDir.Y,
                        oViewDir.Z);
            // position
            Point3D oPos = oCurVP.Position;
            Debug.Print("Position:<{0},{1},{2}> ",
                        oPos.X,
                        oPos.Y,
                        oPos.Z);
            // target
            Point3D oTarget =
                new Point3D(oPos.X + oViewDir.X * oFocal,
                            oPos.Y + oViewDir.Y * oFocal,
                            oPos.Z + oViewDir.Z * oFocal);
            Debug.Print("Target:<{0},{1},{2}> ",
                            oTarget.X,
                            oTarget.Y,
                            oTarget.Z);
 
            // rotation information
            AxisAndAngleResult oAR =
                oRot.ToAxisAndAngle();
            Debug.Print("Rotation Axis:<{0},{1},{2}> ",
                        oAR.Axis.X,
                        oAR.Axis.Y,
                        oAR.Axis.Z);
 
            Debug.Print("Rotation Angle: " +
                oAR.Angle);      
 
        }
        /// <summary>
        /// Multiplies the rotation3d.
        /// by Xiaodong Liang, developer at Autodesk.
        /// </summary>
        /// <param name="r2">The r2.</param>
        /// <param name="r1">The r1.</param>
        /// <returns>Rotation3D.</returns>
        private Rotation3D MultiplyRotation3D(Rotation3D r2,Rotation3D r1)
        {       
            Rotation3D oRot =
                new Rotation3D(r2.D * r1.A + r2.A * r1.D +
                                   r2.B * r1.C - r2.C * r1.B,
                               r2.D * r1.B + r2.B * r1.D +
                                   r2.C * r1.A - r2.A * r1.C,
                               r2.D * r1.C + r2.C * r1.D +
                                  r2.A * r1.B - r2.B * r1.A,
                               r2.D * r1.D - r2.A * r1.A -
                                  r2.B * r1.B - r2.C * r1.C);
 
 
            oRot.Normalize();
            return oRot;
        }
        /// <summary>
        /// Changes the view points properties.
        /// by Xiaodong Liang, developer at Autodesk.
        /// </summary>
        private void changeVPPro()
        {
            Document oDoc =
                Autodesk.Navisworks.Api.Application.ActiveDocument;

            // get current viewpoint
            Viewpoint oCurVP = oDoc.CurrentViewpoint;
            // get copy viewpoint
            Viewpoint oCopyVP = oCurVP.CreateCopy();

            oCopyVP.FarPlaneDistance *= 2;
            oCopyVP.NearPlaneDistance *= 2;
            if (oCopyVP.Tool != Tool.NavigateWalk)
                oCopyVP.Tool = Tool.NavigateWalk;
            oCopyVP.LinearSpeed *= 2;
            oCopyVP.AngularSpeed *= 2;
            if (oCopyVP.Projection != ViewpointProjection.Orthographic)
                oCopyVP.Projection = ViewpointProjection.Orthographic;
            if (oCopyVP.RenderStyle != ViewpointRenderStyle.Wireframe)
                oCopyVP.RenderStyle = ViewpointRenderStyle.Wireframe;

            //update the current viewpoint
            oDoc.CurrentViewpoint.CopyFrom(oCopyVP);
        }
        /// <summary>
        /// Move position of the camera 2 distance along X axis
        /// by Xiaodong Liang, developer at Autodesk.
        /// </summary>
        private void moveCameraPos()
        {
            Document oDoc =
                Autodesk.Navisworks.Api.Application.ActiveDocument;
            // make a copy of current viewpoint
            Viewpoint oCurrVCopy = oDoc.CurrentViewpoint.CreateCopy();
            //step to move
            double step = 2;
            //  create the new position
            Point3D newPos =
                new Point3D(oCurrVCopy.Position.X + step,
                            oCurrVCopy.Position.Y,
                            oCurrVCopy.Position.Z);
            oCurrVCopy.Position = newPos;

            // update current viewpoint
            oDoc.CurrentViewpoint.CopyFrom(oCurrVCopy);

        }
        /// <summary>
        /// Move position of the camera 2 along view direction
        /// by Xiaodong Liang, developer at Autodesk.
        /// </summary>
        private void moveCameraAlongViewDir()
        {
            Document oDoc =
                Autodesk.Navisworks.Api.Application.ActiveDocument;
            // make a copy of current viewpoint
            Viewpoint oCurrVCopy = oDoc.CurrentViewpoint.CreateCopy();
            // get view direction
            Vector3D oViewDir = getViewDir(oCurrVCopy);
            //step to move
            double step = 2;
            //  create the new position
            Point3D newPos =
                new Point3D(oCurrVCopy.Position.X + oViewDir.X * step,
                            oCurrVCopy.Position.Y + oViewDir.Y * step,
                            oCurrVCopy.Position.Z + oViewDir.Z * step);
            oCurrVCopy.Position = newPos;

            // update current viewpoint
            oDoc.CurrentViewpoint.CopyFrom(oCurrVCopy);

        }
        /// <summary>
        /// Changes the camera view dir_ way1.
        /// by Xiaodong Liang, developer at Autodesk.
        /// </summary>
        private void changeCameraViewDir_Way1()
        {
            Document oDoc =
                Autodesk.Navisworks.Api.Application.ActiveDocument;

            // make a copy of current viewpoint
            Viewpoint oCurrVCopy = oDoc.CurrentViewpoint.CreateCopy();
            // Focal Distance
            double oFocal =
                oCurrVCopy.FocalDistance;
            Point3D oPos = oCurrVCopy.Position;
            // get current view direction
            Vector3D oViewDir = getViewDir(oCurrVCopy);

            //current target point (Loot At)
            Point3D oTarget =
                new Point3D(oPos.X + oViewDir.X * oFocal,
                            oPos.Y + oViewDir.Y * oFocal,
                            oPos.Z + oViewDir.Z * oFocal);

            //step to move
            double step = 2;

            // new Point At
            Point3D newPointAt = new Point3D(oTarget.X + step,
                                                oTarget.Y + step,
                                                oTarget.Z + step);

            // set to new target
            oCurrVCopy.PointAt(newPointAt);

            // Note: when you set Point At, the position and focal
            // do not change, while target (Look At) changes, i.e.
            // view direction changes.



            // update current viewpoint
            oDoc.CurrentViewpoint.CopyFrom(oCurrVCopy);

        }
        /// <summary>
        /// Change view direction directly.
        /// Note: will require combination of Position, Target and AlignUp
        /// in this demo: implement LEFT and RIGHT view.
        /// By Xiaodong Liang, developer at Autodesk.
        /// </summary>
        private void changeCameraViewDir_Way2()
        {
            Document oDoc =
                     Autodesk.Navisworks.Api.Application.ActiveDocument;

            // make a copy of current viewpoint
            Viewpoint oCurrVCopy = oDoc.CurrentViewpoint.CreateCopy();

            // Focal Distance
            double oFocal =
                oCurrVCopy.FocalDistance;

            // new target is the center of the model
            Point3D oNewTarget = oDoc.Models[0].RootItem.BoundingBox().Center;

            //new direction is X- >> RIGHT
            Vector3D oNewViewDir = new Vector3D(-1, 0, 0);
            //new direction is X >> LEFT
            //Vector3D oNewViewDir = new Vector3D(1, 0, 0);

            //calculate the new position by the target and focal distance
            Point3D oNewPos = new Point3D(oNewTarget.X - oNewViewDir.X * oFocal,
                oNewTarget.Y - oNewViewDir.Y * oFocal,
                oNewTarget.Z - oNewViewDir.Z * oFocal);

            //set the position
            oCurrVCopy.Position = oNewPos;
            //set the target
            oCurrVCopy.PointAt(oNewTarget);
            //set view direction
            oCurrVCopy.AlignDirection(oNewViewDir);
            // set which direction is up: in this case it is Z+
            oCurrVCopy.AlignUp(new Vector3D(0, 0, 1));

            // update current viewpoint
            oDoc.CurrentViewpoint.CopyFrom(oCurrVCopy);
        }
        /// <summary>
        /// Help function: get view direction.
        /// by Xiaodong Liang, developer at Autodesk.
        /// </summary>
        /// <param name="oVP">The o vp.</param>
        /// <returns>Vector3D.</returns>
        private Vector3D getViewDir(Viewpoint oVP)
        {
            Rotation3D oRot = oVP.Rotation;

            // calculate view direction
            Rotation3D oNegtiveZ =
                new Rotation3D(0, 0, -1, 0);

            Rotation3D otempRot =
                MultiplyRotation3D(oNegtiveZ, oRot.Invert());

            Rotation3D oViewDirRot =
                MultiplyRotation3D(oRot, otempRot);

            // get view direction
            Vector3D oViewDir =
                new Vector3D(oViewDirRot.A,
                            oViewDirRot.B,
                            oViewDirRot.C);

            oViewDir.Normalize();

            return new Vector3D(oViewDir.X,
                        oViewDir.Y,
                        oViewDir.Z);

        }
        /// <summary>
        /// Set up vector
        /// by Xiaodong Liang, developer at Autodesk.
        /// </summary>
        private void changeUp()
        {
            Document oDoc =
                Autodesk.Navisworks.Api.Application.ActiveDocument;

            // make a copy of current viewpoint
            Viewpoint oCurrVCopy = oDoc.CurrentViewpoint.CreateCopy();

            // a new up vector. e.g. Y +
            Vector3D oNewUpVec = new Vector3D(0, 1, 0);

            oCurrVCopy.AlignUp(oNewUpVec);

            // update current viewpoint
            oDoc.CurrentViewpoint.CopyFrom(oCurrVCopy);

        }
        /// <summary>
        /// Rotates the camera.
        /// By Xiaodong Liang, developer at Autodesk.
        /// </summary>
        private void rotateCamera()
        {
            Document oDoc =
                Autodesk.Navisworks.Api.Application.ActiveDocument;

            //  Make a copy of current viewpoint   
            Viewpoint oCurrVCopy =
                oDoc.CurrentViewpoint.CreateCopy();

            //  set the axis we will rotate around （¨Ｚú：o０°,０°,１±）?
            UnitVector3D odeltaA =
                new UnitVector3D(0, 0, 1);

            // Create delta of Quaternion: axis is Z,
            // angle is 45 degree
            Rotation3D delta =
                new Rotation3D(odeltaA, 3.14 / 4);

            // multiply the current Quaternion with the delta ,
            // get the new Quaternion
            oCurrVCopy.Rotation =
                MultiplyRotation3D(delta, oCurrVCopy.Rotation);

            // Update current viewpoint
            oDoc.CurrentViewpoint.CopyFrom(oCurrVCopy);

        }
    }
}
