﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-07-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-08-2014
// ***********************************************************************
// <copyright file="CallBackGeomListener.cs">
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Autodesk.Navisworks.Api;
using ComBridge = Autodesk.Navisworks.Api.ComApi.ComApiBridge;
using COMApi = Autodesk.Navisworks.Api.Interop.ComApi;
using Autodesk.Navisworks.Api.Interop.ComApi;
using DynamicRelaxationClashVerification.Models.DataModels;

/// <summary>
/// The Models namespace includes all the models and interfaces in the MVVM arcitechture. 
/// </summary>
namespace DynamicRelaxationClashVerification.Models
{
    /// <summary>
    /// The CallBackGeomListener class uses the interface InwSimplePrimitivesCB from
    /// Navisworks COM API to get all the primitives from a fragment (model primitiv)
    /// ex. the points that a triangle is made of. 
    /// </summary>
    public class CallBackGeomListener : COMApi.InwSimplePrimitivesCB
    {
        /// <summary>
        /// The current translation vector for the fragment. The COM
        /// API only gives local coord for the primitives. The translation
        /// vector is used to generate global coord. 
        /// </summary>
        private Vector3D translationVector;
        /// <summary>
        /// Sets the translation vector.
        /// </summary>
        /// <param name="translationArray">The translation array.</param>
        public void SetTranslationVector(Vector3D translationArray)
        {
            translationVector = translationArray;
        }
        /// <summary>
        /// Gets or sets a value indicating whether this instance is path.
        /// </summary>
        /// <value><c>true</c> if this instance is path; otherwise, <c>false</c>.</value>
        public bool IsPath { get; set; }
        /// <summary>
        /// Initializes a new instance of the <see cref="CallBackGeomListener"/> class.
        /// </summary>
        /// <param name="isPath">if set to <c>true</c> [is path].</param>
        public CallBackGeomListener(bool isPath)
        {
            IsPath = isPath;
        }
        /// <summary>
        /// The _path
        /// </summary>
        private List<Point3D> _path = new List<Point3D>();
        /// <summary>
        /// The path generated if IsPath is true.
        /// </summary>
        /// <value>The path.</value>
        public List<Point3D> Path
        {
            get { return _path; }
            set
            {
                _path = value;
            }
        }
        /// <summary>
        /// The point list
        /// </summary>
        public List<Point3D> PointList = new List<Point3D>();
        /// <summary>
        /// The path lines
        /// </summary>
        public List<Line3D> PathLines = new List<Line3D>();
        /// <summary>
        /// Method that is called if the fragment is a line. It gives the 
        /// two points (start and end point) of the line as parameters. If
        /// IsPath is true, the points are converted to global coord, before 
        /// added to the path list. 
        /// </summary>
        /// <param name="v1">The v1.</param>
        /// <param name="v2">The v2.</param>
        public void Line(COMApi.InwSimpleVertex v1, COMApi.InwSimpleVertex v2)
        {
            Array array_v1 = (Array)(object)v1.coord;
            Vector3D vc_1 = new Vector3D((float)(array_v1.GetValue(1)), (float)(array_v1.GetValue(2)), (float)(array_v1.GetValue(3)));
            Point3D point_1 = new Point3D(translationVector.X + vc_1.X, translationVector.Y + vc_1.Y, translationVector.Z + vc_1.Z);
            
            if(IsPath)
                Path.Add(point_1);

            Array array_v2 = (Array)(object)v2.coord;
            Vector3D vc_2 = new Vector3D((float)(array_v2.GetValue(1)), (float)(array_v2.GetValue(2)), (float)(array_v2.GetValue(3)));
            Point3D point_2 = new Point3D(translationVector.X + vc_2.X, translationVector.Y + vc_2.Y, translationVector.Z + vc_2.Z);

            if (IsPath)
            {
                Path.Add(point_2);

                PathLines.Add(new Line3D(point_1, point_2));
            }
        }
        /// <summary>
        /// Method that is called if the fragment is a point. It gives the 
        /// the point as parameters. This method is currently not in use, as
        /// no points fragments are needed.
        /// </summary>
        /// <param name="v1">The v1.</param>
        public void Point(COMApi.InwSimpleVertex v1)
        {
            //Do nothing
        }
        /// <summary>
        /// Method that is called if the fragment is a snap point. It gives the 
        /// the point as parameters. This method is currently not in use, as
        /// no points fragments are needed.
        /// </summary>
        /// <param name="v1">The v1.</param>
        public void SnapPoint(COMApi.InwSimpleVertex v1)
        {
            //Do nothing
        }
        /// <summary>
        /// Method that is called if the fragment is a triangle. It gives the 
        /// the three points that the triangle is made of as parameters. It converts 
        /// the points into global coord.
        /// </summary>
        /// <param name="v1">The v1.</param>
        /// <param name="v2">The v2.</param>
        /// <param name="v3">The v3.</param>
        public void Triangle(COMApi.InwSimpleVertex v1, COMApi.InwSimpleVertex v2, COMApi.InwSimpleVertex v3)
        {
            Array array_v1 = (Array)(object)v1.coord;
            Vector3D vc_1 = new Vector3D((float)(array_v1.GetValue(1)), (float)(array_v1.GetValue(2)), (float)(array_v1.GetValue(3)));
            Point3D point_1 = new Point3D(translationVector.X + vc_1.X, translationVector.Y + vc_1.Y, translationVector.Z + vc_1.Z);
            PointList.Add(point_1);

            Array array_v2 = (Array)(object)v2.coord;
            Vector3D vc_2 = new Vector3D((float)(array_v2.GetValue(1)), (float)(array_v2.GetValue(2)), (float)(array_v2.GetValue(3)));
            Point3D point_2 = new Point3D(translationVector.X + vc_2.X, translationVector.Y + vc_2.Y, translationVector.Z + vc_2.Z);
            PointList.Add(point_2);

            Array array_v3 = (Array)(object)v3.coord;
            Vector3D vc_3 = new Vector3D((float)(array_v3.GetValue(1)), (float)(array_v3.GetValue(2)), (float)(array_v3.GetValue(3)));
            Point3D point_3 = new Point3D(translationVector.X + vc_3.X, translationVector.Y + vc_3.Y, translationVector.Z + vc_3.Z);
            PointList.Add(point_3);
        }
    }
}
