﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-17-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-09-2014
// ***********************************************************************
// <copyright file="Clash3D.cs">
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.IO;

//Resources from Autodesk Navisworks .NET and COM API
using Autodesk.Navisworks.Api;
using Autodesk.Navisworks.Api.Clash;
using ComBridge = Autodesk.Navisworks.Api.ComApi.ComApiBridge;
using COMApi = Autodesk.Navisworks.Api.Interop.ComApi;
using Autodesk.Navisworks.Api.Interop.ComApi;

/// <summary>
/// The DataModels namespace includes more primitive model classes, than the
/// Models namespace. It includes classes as Line3D, Clash3D, Path etc. The 
/// similarity with the differens datamodels are that they contains information
/// about data for use in the other models.
/// </summary>
namespace DynamicRelaxationClashVerification.Models.DataModels
{
    /// <summary>
    /// The class Clash3D represents a 3D clash in Navisworks. The
    /// Autodesk.Navisworks.Api.Clash namespace only includes 2D clashes. 
    /// The Clash3D class has 2 properties that represents the two model
    /// objects in Navisworks that have clashed. The Clash class from Navisworks 
    /// API, only includes which triangular surfaces that have clashes. This class takes
    /// all clashes that have the same triangular surfaces in a clash, and creates a 3D clash.
    /// </summary>
    
    public class Clash3D
    {
        /// <summary>
        /// The spesific name of the 3D clash, if needed.
        /// </summary>
        /// <value>The name.</value>
        public string Name { get; set; }
        /// <summary>
        /// The spesific ID of the 3D clash, if needed.
        /// </summary>
        /// <value>The identifier.</value>
        public int ID { get; set; }
        /// <summary>
        /// The 3D clash distance that the two model objects have clashed into
        /// each other. 
        /// </summary>
        /// <value>The distance.</value>
        public double Distance { get; set; }
        /// <summary>
        /// The 3D clash point position (x,y,z) in string format.
        /// </summary>
        /// <value>The clash point in string.</value>
        public string ClashPointInString { get; set; }
        /// <summary>
        /// The 3D point in the path, where the 3D clash occured.
        /// </summary>
        /// <value>The real path point.</value>
        public Point3D RealPathPoint { get; set; }
        /// <summary>
        /// The index of where the 3D point in the path, where the 3D clash occured.
        /// </summary>
        /// <value>The index of the real path point.</value>
        public int RealPathPointIndex { get; set; }
        /// <summary>
        /// The _clash point
        /// </summary>
        private Point3D _clashPoint;
        /// <summary>
        /// The 3D clash point where the clash occur.
        /// </summary>
        /// <value>The clash point.</value>
        public Point3D ClashPoint
        {
            get 
            { 
                return _clashPoint; 
            }
            set
            {
                _clashPoint = value;
                SetClashPointString(value);
            }
        }
        /// <summary>
        /// Method that sets the clash point into string.
        /// </summary>
        /// <param name="point">The point.</param>
        private void SetClashPointString(Point3D point)
        {
            ClashPointInString = "(" + point.X + ", " + point.Y + ", " + point.Z + ")";
        }
        /// <summary>
        /// The selection A is the validation object in the 3D clash. 
        /// </summary>
        /// <value>The selection A.</value>
        public ModelItemCollection SelectionA { get; set; }
        /// <summary>
        /// The selection B is the boundary geometry in the 3D clash.
        /// </summary>
        /// <value>The selection b.</value>
        public ModelItemCollection SelectionB { get; set; }
        /// <summary>
        /// Method that calculates and returns the clash direction vector. 
        /// The clash direction vector is the normal vector out from the
        /// boundary geometry surface. It is used to find which direction,
        /// the validation object should move in DR. 
        /// 
        /// The method uses Autodesk Navisworks COM API. See seperate 
        /// documentation from Autodesk about the COM API. 
        /// </summary>
        /// <returns>Vector3D.</returns>
        public Vector3D GetClashDirectionVector()
        {
            //convert to COM selection
            COMApi.InwOpState oState = ComBridge.State;
            COMApi.InwOpSelection oSel = ComBridge.ToInwOpSelection(SelectionB);

            // create the callback object. the false parameter is only to tell the callback object that 
            // it is not used to generate a path. 
            CallBackGeomListener callbkListener = new CallBackGeomListener(false);

            foreach (COMApi.InwOaPath3 path in oSel.Paths())
            {
                foreach (COMApi.InwOaFragment3 frag in path.Fragments())
                {
                    //Getting the transformation matrix. See report for more information.
                    COMApi.InwLTransform3f3 localToWorld = (COMApi.InwLTransform3f3)(object)frag.GetLocalToWorldMatrix();
                    Array array_v1 = (Array)(object)localToWorld.Matrix;

                    //The translation values can be found in indices 13, 14 and 15.
                    //
                    // | [ 1 0 0 ]   [x] |
                    // | [ 0 1 0 ]   [y] |      | R  S |
                    // | [ 0 0 1 ]   [z] |  =   | P  1 |
                    // |                 |
                    // | [ 0 0 0 ]   [1] |
                    //
                    // Above is the transformation matrix, where the S matrix is the translation matrix.
                    // See master thesis for more information about the transformation matrix (incl. R, P etc)

                    double x = (Double)array_v1.GetValue(13);
                    double y = (Double)array_v1.GetValue(14);
                    double z = (Double)array_v1.GetValue(15);

                    Vector3D S = new Vector3D(x, y, z);
                    // Setting the translation vector, to be used to generate the primitives.
                    callbkListener.SetTranslationVector(S);
                    frag.GenerateSimplePrimitives(COMApi.nwEVertexProperty.eNORMAL, callbkListener);
                }
            }

            //Calculating vectors, based on triangle points computed from COM API.

            Point3D point_1 = callbkListener.PointList.ElementAt(0);
            Point3D point_2 = callbkListener.PointList.ElementAt(1);
            Point3D point_3 = callbkListener.PointList.ElementAt(2);

            Vector3D A = new Vector3D(point_2.X - point_1.X, point_2.Y - point_1.Y, point_2.Z - point_1.Z);
            Vector3D B = new Vector3D(point_3.X - point_1.X, point_3.Y - point_1.Y, point_3.Z - point_1.Z);

            //Using cross product to find the normal vector out for the plane given by the two vectors; A and B
            Vector3D res = B.Cross(A);
 
            return res;
        }

    }
}
