﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-17-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-08-2014
// ***********************************************************************
// <copyright file="Path.cs" company="">
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Autodesk.Navisworks.Api;
using ComBridge = Autodesk.Navisworks.Api.ComApi.ComApiBridge;
using COMApi = Autodesk.Navisworks.Api.Interop.ComApi;
using Autodesk.Navisworks.Api.Interop.ComApi;
using System.Diagnostics;

/// <summary>
/// The DataModels namespace includes more primitive model classes, than the
/// Models namespace. It includes classes as Line3D, Clash3D, Path etc. The 
/// similarity with the differens datamodels are that they contains information
/// about data for use in the other models.
/// </summary>
namespace DynamicRelaxationClashVerification.Models.DataModels
{
    /// <summary>
    /// The Path class includes data and methods for creating, optimizing and store
    /// information about the path. The class' path factory, uses Navisworks COM API
    /// together with a sorting method, to generate the path. The path is represented 
    /// by a list with Point3D.
    /// </summary>
    public class Path : ObservableProgress
    {      
        /// <summary>
        /// A list of all the nodes that defines the path.
        /// Is filled using the RunPathFactory
        /// </summary>
        private List<Point3D> _currentPath = new List<Point3D>();
        /// <summary>
        /// Gets or sets the current path.
        /// </summary>
        /// <value>The current path.</value>
        public List<Point3D> CurrentPath 
        { 
            get {return _currentPath;}
            set
            {
                _currentPath = value;
            }
        }
        /// <summary>
        /// Method that creates a list of Point3D, of a 
        /// ModelItemCollection input using Navisworks COM API.
        /// </summary>
        /// <param name="oModelColl">ModelItemCollection</param>
        public void RunPathFactory(ModelItemCollection oModelColl)
        {
            if (oModelColl == null)
                return;

            //convert to COM selection
            COMApi.InwOpState oState = ComBridge.State;
            COMApi.InwOpSelection oSel = ComBridge.ToInwOpSelection(oModelColl);

            // create the callback object. the false parameter is only to tell the callback object that 
            // it is not used to generate a path.
            CallBackGeomListener callbkListener = new CallBackGeomListener(true);

            foreach (COMApi.InwOaPath3 path in oSel.Paths())
            {
                foreach (COMApi.InwOaFragment3 frag in path.Fragments())
                {
                    //Getting the transformation matrix. See report for more information.
                    COMApi.InwLTransform3f3 localToWorld = (COMApi.InwLTransform3f3)(object)frag.GetLocalToWorldMatrix();

                    //The translation values can be found in indices 13, 14 and 15.
                    //
                    // | [ 1 0 0 ]   [x] |
                    // | [ 0 1 0 ]   [y] |      | R  S |
                    // | [ 0 0 1 ]   [z] |  =   | P  1 |
                    // |                 |
                    // | [ 0 0 0 ]   [1] |
                    //
                    // Above is the transformation matrix, where the S matrix is the translation matrix.
                    // See master thesis for more information about the transformation matrix (incl. R, P etc)

                    Array array_v1 = (Array)(object)localToWorld.Matrix;
                    double x = (Double)array_v1.GetValue(13);
                    double y = (Double)array_v1.GetValue(14);
                    double z = (Double)array_v1.GetValue(15);

                    Vector3D S = new Vector3D(x, y, z);
                    // Setting the translation vector, to be used to generate the primitives.
                    callbkListener.SetTranslationVector(S);
                    frag.GenerateSimplePrimitives(COMApi.nwEVertexProperty.eNORMAL, callbkListener);
                }
            }
            //Setting the current path of this object, to be the generated path from callbacklistener. (COM API)
            this.CurrentPath = callbkListener.Path;

            //As the list of path points not are sorted (are in random order), it needs to be sorted.
            SortCurrentPath(callbkListener.PathLines);
            //Because of the way that lines are created in Navisworks (as line segments),
            //must the duplication nodes be removed. There will be a duplication in each corner in the path.
            RemoveDuplicationsInPath();
        }
        /// <summary>
        /// The method sorts a list of line segments, so that they occur in
        /// correct order. This is done by looking at which line segments that
        /// have equal line segment points. As Navisworks API doesn't give totally 
        /// equal points, a point tolerance is set to 5 mm. If two points have a distance 
        /// smaller than 5 mm between them, they are seen as the same point. 
        /// 
        /// It returns a list of Point3D (nodes).
        /// </summary>
        /// <param name="lineList">The line segment list.</param>
        public void SortCurrentPath(List<Line3D> lineList)
        {
            List<Point3D> sortedList = new List<Point3D>();
            int length = CurrentPath.Count();

            sortedList.Add(lineList.ElementAt(0).V1);
            sortedList.Add(lineList.ElementAt(0).V2);

            lineList.RemoveAt(0);

            int counter = 0; //Hack for Aker Models
            while (sortedList.Count() != length && lineList.Count() > 0)
            {
                if (lineList.Count <= 2)
                {
                    counter++;
                }
                if (counter > 5)
                {
                    break;
                }
                ProgressMin = 0;
                ProgressMax = lineList.Count();

                for (int i = 0; i < lineList.Count(); i++)
                {
                    if ((Math.Abs(sortedList.Last().X - lineList.ElementAt(i).V1.X) < 5) &&
                        (Math.Abs(sortedList.Last().Y - lineList.ElementAt(i).V1.Y) < 5) &&
                        (Math.Abs(sortedList.Last().Z - lineList.ElementAt(i).V1.Z) < 5))
                    {
                        sortedList.Add(lineList.ElementAt(i).V1);
                        sortedList.Add(lineList.ElementAt(i).V2);

                        lineList.RemoveAt(i);
                        FireProgressChanged(i);
                        continue;
                    }
                    else if ((Math.Abs(sortedList.Last().X - lineList.ElementAt(i).V2.X) < 5) &&
                             (Math.Abs(sortedList.Last().Y - lineList.ElementAt(i).V2.Y) < 5) &&
                             (Math.Abs(sortedList.Last().Z - lineList.ElementAt(i).V2.Z) < 5))
                    {
                        sortedList.Add(lineList.ElementAt(i).V2);
                        sortedList.Add(lineList.ElementAt(i).V1);

                        lineList.RemoveAt(i);
                        FireProgressChanged(i);
                        continue;
                    }

                    else if ((Math.Abs(sortedList.First().X - lineList.ElementAt(i).V1.X) < 5) &&
                             (Math.Abs(sortedList.First().Y - lineList.ElementAt(i).V1.Y) < 5) &&
                             (Math.Abs(sortedList.First().Z - lineList.ElementAt(i).V1.Z) < 5))
                    {
                        sortedList.Insert(0,lineList.ElementAt(i).V1);
                        sortedList.Insert(0,lineList.ElementAt(i).V2);

                        lineList.RemoveAt(i);
                        FireProgressChanged(i);
                        continue;
                    }
                    else if ((Math.Abs(sortedList.First().X - lineList.ElementAt(i).V2.X) < 5) &&
                             (Math.Abs(sortedList.First().Y - lineList.ElementAt(i).V2.Y) < 5) &&
                             (Math.Abs(sortedList.First().Z - lineList.ElementAt(i).V2.Z) < 5))
                    {
                        sortedList.Insert(0,lineList.ElementAt(i).V2);
                        sortedList.Insert(0,lineList.ElementAt(i).V1);

                        lineList.RemoveAt(i);
                        FireProgressChanged(i);
                        continue;
                    }
                }
            }
            ProgressMin = 0;
            ProgressMax = 100;
            FireProgressChanged(100);
            this.CurrentPath = sortedList;
        }
        /// <summary>
        /// Removes the duplications in path.
        /// </summary>
        public void RemoveDuplicationsInPath()
        {
            for (int i = 0; i < CurrentPath.Count() - 1; i++)
            {
                if (CurrentPath.ElementAt(i).X - CurrentPath.ElementAt(i + 1).X < 5 && 
                    CurrentPath.ElementAt(i).Y - CurrentPath.ElementAt(i + 1).Y < 5)
                {
                    CurrentPath.RemoveAt(i + 1);
                }
            }
        }
        /// <summary>
        /// Prints the path to file.
        /// </summary>
        public void PrintPathToFile()
        {
            // Compose a string that consists of three lines.
            string lines = "";

            // Write the string to a file.
            System.IO.StreamWriter file = new System.IO.StreamWriter("C:\\Master Thesis/test.txt");

            foreach(Point3D point in CurrentPath)
            {
                lines += point.X + "\n";
            }

            lines += "\n\n\n";

            foreach (Point3D point in CurrentPath)
            {
                lines += point.Y + "\n";
            }

            lines += "\n\n\n";

            foreach (Point3D point in CurrentPath)
            {
                lines += point.Z + "\n";
            }
            file.WriteLine(lines);

            file.Close();
        }
        /// <summary>
        /// Runs the path optimalization. The optimalization uses the clashes from the
        /// clashlist to move the points in the realpath. 
        /// </summary>
        /// <param name="realPath">The real path.</param>
        /// <param name="clashList">The clash list.</param>
        /// <returns>List&lt;Point3D&gt;.</returns>
        public List<Point3D> RunPathOptimalization(List<Point3D> realPath, List<Clash3D> clashList)
        {
            foreach (Clash3D clash in clashList)
            {
                Point3D clashPoint = clash.ClashPoint;
                Point3D pathPoint = realPath.ElementAt(clash.RealPathPointIndex);

                Point3D prePoint = realPath.ElementAt(clash.RealPathPointIndex - 1);
                Point3D nextPoint = realPath.ElementAt(clash.RealPathPointIndex + 1);

                Point3D prePrePoint = realPath.ElementAt(clash.RealPathPointIndex - 2);
                Point3D nextNextPoint = realPath.ElementAt(clash.RealPathPointIndex + 2);
                
                double clashDistance = Math.Abs(clash.Distance);

                Vector3D vc = clash.GetClashDirectionVector();

                UnitVector3D uVc = new UnitVector3D(vc);

                Point3D newPoint = new Point3D(
                    pathPoint.X + uVc.X*clashDistance,
                    pathPoint.Y + uVc.Y*clashDistance,
                    pathPoint.Z + uVc.Z*clashDistance);

                Point3D newPrePoint = new Point3D(
                    prePoint.X + uVc.X * clashDistance * 0.5,
                    prePoint.Y + uVc.Y * clashDistance * 0.5,
                    prePoint.Z + uVc.Z * clashDistance * 0.5);

                Point3D newNextPoint = new Point3D(
                    nextPoint.X + uVc.X * clashDistance * 0.5,
                    nextPoint.Y + uVc.Y * clashDistance * 0.5,
                    nextPoint.Z + uVc.Z * clashDistance * 0.5);

                Point3D newPrePrePoint = new Point3D(
                    prePrePoint.X + uVc.X * clashDistance * 0.25,
                    prePrePoint.Y + uVc.Y * clashDistance * 0.25,
                    prePrePoint.Z + uVc.Z * clashDistance * 0.25);

                Point3D newNextNextPoint = new Point3D(
                    nextNextPoint.X + uVc.X * clashDistance * 0.25,
                    nextNextPoint.Y + uVc.Y * clashDistance * 0.25,
                    nextNextPoint.Z + uVc.Z * clashDistance * 0.25);

                realPath.Insert(clash.RealPathPointIndex + 2, nextNextPoint);
                realPath.RemoveAt(clash.RealPathPointIndex + 2 + 1);

                realPath.Insert(clash.RealPathPointIndex + 1, nextPoint);
                realPath.RemoveAt(clash.RealPathPointIndex + 1 + 1);

                realPath.Insert(clash.RealPathPointIndex, newPoint);
                realPath.RemoveAt(clash.RealPathPointIndex + 1);

                realPath.Insert(clash.RealPathPointIndex-1, prePoint);
                realPath.RemoveAt(clash.RealPathPointIndex-1 + 1);

                realPath.Insert(clash.RealPathPointIndex - 2, prePrePoint);
                realPath.RemoveAt(clash.RealPathPointIndex - 2 + 1);
            }

            return realPath;
        }
        /// <summary>
        /// Reverts the start position of the validation object
        /// </summary>
        public void RevertStartPosition()
        {
            CurrentPath.Reverse();
        }
    }

}
