﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-17-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-09-2014
// ***********************************************************************
// <copyright file="ValidationObject.cs">
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Autodesk.Navisworks.Api;
using ComBridge = Autodesk.Navisworks.Api.ComApi.ComApiBridge;
using COMApi = Autodesk.Navisworks.Api.Interop.ComApi;
using Autodesk.Navisworks.Api.Interop.ComApi;

/// <summary>
/// The DataModels namespace includes more primitive model classes, than the
/// Models namespace. It includes classes as Line3D, Clash3D, Path etc. The 
/// similarity with the differens datamodels are that they contains information
/// about data for use in the other models.
/// </summary>
namespace DynamicRelaxationClashVerification.Models.DataModels
{
    /// <summary>
    /// The ValidationObject class contains data and mathods handling the validation 
    /// object. The class has properties needed for calculations in dynamic relaxation
    /// algorithm, and method for moving and rotating tha validation object in Navisworks,
    /// through the Navisworks .NET and COM API. 
    /// </summary>
    public class ValidationObject
    {
        /// <summary>
        /// The current Navisworks document opened in NW.
        /// </summary>
        protected Document doc = Autodesk.Navisworks.Api.Application.MainDocument;
        /// <summary>
        /// The validation object represented as a ModelItemCollection
        /// from Navisworks API. 
        /// </summary>
        public ModelItemCollection Model;

        /*
         * The following properties are dynamic relaxation properties
         * used in the Dynamic Relaxation Solver.
         */

        /// <summary>
        /// The mass of the validation object.
        /// </summary>
        /// <value>2.2 (static, should not be changed)</value>
        public double Mass { get; set; }
        /// <summary>
        /// The stiffness of the validation object. More info about stiffness, 
        /// see master thesis. 
        /// </summary>
        /// <value>1 (static, should not be changed)</value>
        public double Stiffness { get; set; }
        /// <summary>
        /// The applied force on the validation object during clash.
        /// Elements: [x,y,z]
        /// </summary>
        /// <value>The force vector</value>
        public double[] Force { get; set; }
        /// <summary>
        /// The velocity of the validation object during dynamic relaxation.
        /// Elements: [x, y, z]
        /// </summary>
        /// <value>The velocity vector.</value>
        public double[] Velocity { get; set; }
        /// <summary>
        /// The acceleration of the validation object during dynamic relaxation.
        /// Elements: [x, y, z]
        /// </summary>
        /// <value>The acceleration vector</value>
        public double[] Acceleration { get; set; }
        /// <summary>
        /// The damping coefficient for the validation object in the dynamic relaxation
        /// equations. 
        /// </summary>
        /// <value>The damping coefficient.</value>
        public double Damping { get; set; }

        /*
         * Axle properties used in cases where the
         * validation object is a trolley, or a simular 
         * transportation unit with axle constraints.
         */ 

        /// <summary>
        /// The axle position of the moveable axle. 
        /// Can be front, rear or center. Center is default, if nothing
        /// else are selected. Use center if validation object is ex. a stretcher.
        /// </summary>
        /// <value>The axle position.</value>
        public AxlePosition AxlePosition { get; set; }
        /// <summary>
        /// The distance between the axle and the center of
        /// validation object. Is set to zero as default.
        /// </summary>
        /// <value>The axle center distance.</value>
        public double AxleCenterDistance { get; set; }
        /// <summary>
        /// Initializes a new instance of the <see cref="ValidationObject"/> class.
        /// </summary>
        /// <param name="model">The model.</param>
        public ValidationObject(ModelItemCollection model)
        {
            this.Model = model;
            AxleCenterDistance = 0;
            AxlePosition = AxlePosition.Center;
        }
        /// <summary>
        /// Method that sets the position of the validation object,
        /// given the point to move too (startPoint), and the rotation 
        /// needed to get the correct direction of the validation object.
        /// </summary>
        /// <param name="startPoint">The start point.</param>
        /// <param name="rot">The rotation .</param>
        /// <exception cref="System.Exception"></exception>
        public void SetValidationObjectPosition(Point3D startPoint, Rotation3D rot)
        {
            //Global center points
            double center_x = Model.BoundingBox().Center.X;
            double center_y = Model.BoundingBox().Center.Y;

            //Setting the correct center point based on the axle position.
            //The axle mid point is used as the center position, to get a more
            //realistic movement of validation object.
            if (AxlePosition == AxlePosition.Front)
                center_y = Model.BoundingBox().Center.Y + AxleCenterDistance;
            else if (AxlePosition == AxlePosition.Rear)
                center_y = Model.BoundingBox().Center.Y - AxleCenterDistance;
            else
                center_y = Model.BoundingBox().Center.Y;

            Vector3D center = new Vector3D(center_x, center_y, 0);

            //Move validation object to origo (because of NW API)
            MoveValidationObject(new Vector3D(center.X*-1, center.Y*-1, 0));

            //Rotate to correct direction
            RotateValidationObject(rot, new Point3D(0, 0, 0));

            //Move to startpoint
            MoveValidationObject(new Vector3D(startPoint.ToVector3D().X, startPoint.ToVector3D().Y, 0));
        }
        /// <summary>
        /// Method for moving the Model item in the
        /// direction given by the 3D vector. 
        /// </summary>
        /// <param name="newDirection">The new direction.</param>
        /// <exception cref="System.Exception"></exception>
        public void MoveValidationObject(Vector3D newDirection)
        {
            if (Model.Count == 0)
                throw new Exception();

            //Creates a transformation matrix, based on the translation vector.
            Transform3D oNewOverrideTrans = Transform3D.CreateTranslation(newDirection);

            //Overrides the current position sith the new transform. See NW doc
            //to read why you need to override permanent, and what this mean.
            doc.Models.OverridePermanentTransform(Model, oNewOverrideTrans, true);
            
        }
        /// <summary>
        /// Rotates the validation object, based on a rotation (angle) and 
        /// a rotation point, usually the center of the validation object. 
        /// </summary>
        /// <param name="rotation">The rotation.</param>
        /// <param name="rotationPoint">The rotation point.</param>
        public void RotateValidationObject(Rotation3D rotation, Point3D rotationPoint)
        {
            //Creates a opposite vector based on the position, to 
            //move the validation object to origo, for rotation.
            Vector3D vec = rotationPoint.ToVector3D().Multiply(-1.0);

            //Move to origo
            MoveValidationObject(vec);

            //Do transformation, and override transform.
            Transform3D oNewOverrideRotationTrans = new Transform3D(rotation);
            doc.Models.OverridePermanentTransform(Model, oNewOverrideRotationTrans, true);

            //Move back to original position.
            MoveValidationObject(vec.Multiply(-1.0));
        }
        /// <summary>
        /// Initializes the forces and moments.
        /// </summary>
        public void InitForcesAndMoments()
        {
            Force = new double[] { 0, 0, 0 };
        }
        /// <summary>
        /// Initializes all dynamic properties.
        /// </summary>
        public void InitAllDynamicProperties()
        {
            Force = new double[] { 0, 0, 0 };
            Velocity = new double[] { 0, 0, 0 };
            Acceleration = new double[] { 0, 0, 0 };
        }
    }
}
