﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-07-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-09-2014
// ***********************************************************************
// <copyright file="DynamicClashTest.cs" company="">
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Autodesk.Navisworks.Api;
using Autodesk.Navisworks.Api.Clash;
using Autodesk.Navisworks.Api.DocumentParts;
using System.Collections.ObjectModel;
using DynamicRelaxationClashVerification.Models.DataModels;


/// <summary>
/// The Models namespace includes all the models and interfaces in the MVVM arcitechture. 
/// </summary>
namespace DynamicRelaxationClashVerification.Models
{
    /// <summary>
    /// The DynamicClashTest class contains properties and methods for creating and running clash tests 
    /// through the Navisworks .NET API. Instead of using Navisworks formatting for clashes, a clash group
    /// containing the same composite items (clash model items), will be used as ONE 3D clash. This means that:
    /// 1 NW group = 1 3D Clash, 1 NW clash test = 1 3D clash test iteration.
    /// </summary>
    public class DynamicClashTest
    {
        /// <summary>
        /// The current document for use with the Navisworks API
        /// </summary>
        private Document _document;
        /// <summary>
        /// The current clash document in Navisworks
        /// </summary>
        private DocumentClash _documentClash;
        /// <summary>
        /// The current clash tests in current clash document
        /// </summary>
        private DocumentClashTests _oDCT;
        /// <summary>
        /// List of all 3D clashes that have occured in the clash testing.
        /// </summary>
        private List<Clash3D> _clashes = new List<Clash3D>();
        /// <summary>
        /// Current validation object as ModelItemCollection
        /// </summary>
        /// <value>The validation object.</value>
        private ModelItemCollection ValidationObject { get; set; }
        /// <summary>
        /// Current boundary geometry as ModelItemCollection
        /// </summary>
        /// <value>The boundary geometry.</value>
        private ModelItemCollection BoundaryGeometry { get; set; }
        /// <summary>
        /// The clash type (hard, hard conservativ, etc)
        /// </summary>
        /// <value>The type of the clash.</value>
        public ClashTestType ClashType { get; set; }
        /// <summary>
        /// The clash tolerance.
        /// </summary>
        /// <value>The clash tolerance.</value>
        public double ClashTolerance { get; set; }
        /// <summary>
        /// Initializes a new instance of the <see cref="DynamicClashTest"/> class.
        /// </summary>
        /// <param name="validationObject">The validation object.</param>
        /// <param name="boundaryGeometry">The boundary geometry.</param>
        public DynamicClashTest(ModelItemCollection validationObject, ModelItemCollection boundaryGeometry)
        {
            //Setting the document to the active (opened) document/file in Navisworks
            _document = Autodesk.Navisworks.Api.Application.ActiveDocument;
            //Getting the Clash document from Navisworks.
            _documentClash = _document.GetClash();
            //Creating a clash test document for all the test data in Navisworks.
            _oDCT = _documentClash.TestsData;

            ValidationObject = validationObject;
            BoundaryGeometry = boundaryGeometry;

            ClashCollection = new ObservableCollection<Clash3D>();
            ClashCollectionList = new ObservableCollection<Clash3D>();
        }
        /// <summary>
        /// Creates a new dynamic clash test, and setting the
        /// necessary properties.
        /// </summary>
        public void CreateDynamicClashTest()
        {
            ClashTest newClashTest = new ClashTest();
            newClashTest.DisplayName = "DR Verification Clash Test";

            //Set SelectionA and SelectionB of the ClashTest
            ModelItemCollection oSelA = ValidationObject;
            ModelItemCollection oSelB = BoundaryGeometry;

            newClashTest.SelectionA.Selection.CopyFrom(oSelA);
            newClashTest.SelectionB.Selection.CopyFrom(oSelB);
            //set other properties of the test …-}

            newClashTest.Tolerance = ClashTolerance;
            newClashTest.TestType = ClashType;
            newClashTest.Status = ClashTestStatus.New;

            _oDCT.TestsAddCopy(newClashTest);

            ClashCollection.Clear();
        }
        /// <summary>
        /// Gets or sets a value indicating whether this instance is clashes.
        /// </summary>
        /// <value><c>true</c> if this instance is clashes; otherwise, <c>false</c>.</value>
        public bool IsClashes { get; set; }
        /// <summary>
        /// Runs the spesific ClashTest in Navisworks. Saves and creates 3D 
        /// clashes, and deletes the clashes tests in navisworks with no
        /// clashes.
        /// </summary>
        public void RunClashTest()
        {
            IsClashes = false;

            int testIndex;
            if (_oDCT.Tests.Count() == 0)
            {
                testIndex = 0;
            }
            else
            {
                testIndex = _oDCT.Tests.IndexOf(_oDCT.Tests.Last());
            }

            ClashTest oFirstTest = _oDCT.Tests[testIndex] as ClashTest;
            _oDCT.TestsRunTest(oFirstTest);

            if ((_oDCT.Tests[testIndex] as ClashTest).Children.Count() == 0)
            {
                _oDCT.TestsRemoveAt(testIndex);
            }
            else
            {
                CreateAndAddClash3D((_oDCT.Tests[testIndex] as ClashTest));
                IsClashes = true;
            }
        }
        /// <summary>
        /// The clash collection with all logged 3D clashes.
        /// </summary>
        /// <value>The clash collection.</value>
        public ObservableCollection <Clash3D> ClashCollection { get; set; }
        /// <summary>
        /// Gets or sets the clash collection list. Contains all points which has been gone through during the dynamic clash test.
        /// </summary>
        /// <value>The clash collection list.</value>
        public ObservableCollection <Clash3D> ClashCollectionList { get; set; }
        /// <summary>
        /// Creates the and add 3D clash to Clash collection.
        /// </summary>
        /// <param name="clashTest">The clash test.</param>
        public void CreateAndAddClash3D(ClashTest clashTest)
        {
            List<List<SavedItem>> clashGroups = CreateClashGroupsBasedOnCompositeItem(clashTest);

            foreach (List<SavedItem> group in clashGroups)
            {
                Clash3D newClash = new Clash3D();
                double distance = 0;
                double counter = 0;

                double x = 0;
                double y = 0;
                double z = 0;

                newClash.SelectionA = new ModelItemCollection();
                newClash.SelectionB = new ModelItemCollection();

                foreach (SavedItem child in group)
                {
                    ClashResult res = child as ClashResult;
                    distance += res.Distance;
                    x += res.Center.X;
                    y += res.Center.Y;
                    z += res.Center.Z;
                    counter++;

                    newClash.SelectionA.Add(res.CompositeItem1);
                    newClash.SelectionB.Add(res.CompositeItem2);
                }
                distance = distance / counter;

                x = x / counter;
                y = y / counter;
                z = z / counter;

                newClash.Distance = distance;
                newClash.ID = ClashCollection.Count();
                newClash.Name = "3D DR Clash #" + ClashCollection.Count();
                newClash.ClashPoint = new Point3D(x, y, z);

                ClashCollectionList.Add(newClash);
                ClashCollection.Add(newClash);
            }
        }
        /// <summary>
        /// Creates the clash groups based on the composite items.
        /// It creates clash groups for each different surface (composite item)
        /// that the validation object clashes with.
        /// </summary>
        /// <param name="clashTest">The clash test.</param>
        /// <returns>List&lt;List&lt;SavedItem&gt;&gt;.</returns>
        private List<List<SavedItem>> CreateClashGroupsBasedOnCompositeItem(ClashTest clashTest)
        {
            List<List<SavedItem>> clashGroups = new List<List<SavedItem>>();
            List<SavedItem> newGroup = new List<SavedItem>();
            newGroup.Add(clashTest.Children.First());
            clashGroups.Add(newGroup);

            int i = 0;
            foreach (SavedItem child in clashTest.Children)
            {
                if (i == 0)
                    continue;
                
                ClashResult res = child as ClashResult;
                bool isFound = false;
                foreach (List<SavedItem> group in clashGroups)
                {
                    if(res.CompositeItem2.Equals((group.First() as ClashResult).CompositeItem2))
                    {
                        group.Add(child);
                        isFound = true;
                        break;
                    }
                }
                if (!isFound)
                {
                    List<SavedItem> newClashGroup = new List<SavedItem>();
                    newClashGroup.Add(child);
                    clashGroups.Add(newClashGroup);
                }
            }
            return clashGroups;
        }
    }
}
