﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-31-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-17-2014
// ***********************************************************************
// <copyright file="DynamicClashVerification.cs" >
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Autodesk.Navisworks.Api;
using System.Collections.ObjectModel;
using Autodesk.Navisworks.Api.Plugins;
using Autodesk.Navisworks.Api.Clash;

using System.Runtime.InteropServices;
using System.Windows.Threading;
using DynamicRelaxationClashVerification.Models.DataModels;

/// <summary>
/// The Models namespace includes all the models and interfaces in the MVVM arcitechture.
/// </summary>
namespace DynamicRelaxationClashVerification.Models
{
    /// <summary>
    /// The DynamicClashVerification class is the "connection"/ controller between the ViewModels and the Models.
    /// It controls which intances and methods that have to be used/ created in the different phases of the
    /// verification. The class also contain information about all necessary data needed to create path, run
    /// optimalization and verification. Each instance of the DynamicClashVerification is its own verification,
    /// and can have different validation object, geometry boundary, path and other settings.
    /// </summary>
    public class DynamicClashVerification 
    {
        /// <summary>
        /// The name of the verification analysis.
        /// </summary>
        /// <value>The name.</value>
        public string Name { get; set; }
        /// <summary>
        /// The description of the verification analysis.
        /// </summary>
        /// <value>The description.</value>
        public string Description { get; set; }
        /// <summary>
        /// The _status
        /// </summary>
        private Status _status = Status.New;
        /// <summary>
        /// Gets or sets the status.
        /// </summary>
        /// <value>The status.</value>
        public Status Status {
            get { return _status; }
            set
            {
                _status = value;
                ChangeStatusIcon();
            }
        }
        /// <summary>
        /// Initializes a new instance of the <see cref="DynamicClashVerification" /> class.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        public DynamicClashVerification(string name, string description)
        {
            Name = name;
            Description = description;            
            Status = Status.New;
        }
        /// <summary>
        /// The selected boundary geometry from the selection sets.
        /// </summary>
        /// <value>The boundary geometry item.</value>
        public ModelItemCollection BoundaryGeometryItem { get; set; }
        /// <summary>
        /// The selected validation object from the selection sets.
        /// </summary>
        /// <value>The validation object item.</value>
        public ModelItemCollection ValidationObjectItem { get; set; }
        /// <summary>
        /// The selected path from the selection sets.
        /// </summary>
        /// <value>The path item.</value>
        public ModelItemCollection PathItem { get; set; }
        /// <summary>
        /// The Dynamic Clash test used to run clash tests.
        /// </summary>
        /// <value>The dynamic clash test.</value>
        public DynamicClashTest DynamicClashTest { get; set; }
        /// <summary>
        /// The dynamic movement instance used to create dynamic movments
        /// for the validation object.
        /// </summary>
        /// <value>The dynamic movement.</value>
        public DynamicMovement DynamicMovement { get; set; }
        /// <summary>
        /// The dynamic relaxation solver instance for solving the
        /// verification using dynamic relaxation algorithm.
        /// </summary>
        /// <value>The dynamic relaxation solver.</value>
        public DynamicRelaxationSolver DynamicRelaxationSolver { get; set; }
        /// <summary>
        /// Gets or sets the movement precision.
        /// </summary>
        /// <value>The movement precision.</value>
        public int MovementPrecision { get; set; }
        /// <summary>
        /// Gets or sets the type of the clash.
        /// </summary>
        /// <value>The type of the clash.</value>
        public ClashTestType ClashType { get; set; }
        /// <summary>
        /// Gets or sets the clash tolerance.
        /// </summary>
        /// <value>The clash tolerance.</value>
        public double ClashTolerance { get; set; }
        /// <summary>
        /// The Path instance for generating and optimizing
        /// the current path.
        /// </summary>
        /// <value>The path.</value>
        public Path Path { get; set; }
        /// <summary>
        /// Gets or sets the validation object.
        /// </summary>
        /// <value>The validation object.</value>
        public ValidationObject ValidationObject { get; set; }
        /// <summary>
        /// Creates/ sets a new path, add a progress listener and
        /// run the path factory (generating the path from COM API).
        /// </summary>
        /// <param name="pathItem">The path item.</param>
        public void SetPath(ModelItemCollection pathItem)
        {
            Path = new Path();
            PathItem = pathItem;
            Path.AddProgress(ProgressListener);
            Path.RunPathFactory(pathItem);
        }
        /// <summary>
        /// Sets the validation object, with its needed properties.
        /// </summary>
        /// <param name="validationObjectItem">The validation object item.</param>
        /// <param name="axleCenterLength">Length of the axle center.</param>
        /// <param name="axlePosition">The axle position.</param>
        public void SetValidationObject(ModelItemCollection validationObjectItem, double axleCenterLength, AxlePosition axlePosition)
        {
            ValidationObject = new ValidationObject(validationObjectItem);
            ValidationObjectItem = validationObjectItem;

            ValidationObject.AxleCenterDistance = axleCenterLength;
            ValidationObject.AxlePosition = axlePosition;
        }
        /// <summary>
        /// Sets the boundary geometry.
        /// </summary>
        /// <param name="boundaryGeometryItem">The boundary geometry item.</param>
        public void SetBoundaryGeometry(ModelItemCollection boundaryGeometryItem)
        {
            BoundaryGeometryItem = boundaryGeometryItem;
        }
        /// <summary>
        /// The progress listener for listning on obervable progress objects;
        /// DynamicRelaxationSolver and DynamicMovement.
        /// </summary>
        /// <value>The progress listener.</value>
        public ProgressListener ProgressListener { get; set; }
        /// <summary>
        /// Gets or sets the result listener.
        /// </summary>
        /// <value>The result listener.</value>
        public ResultListener ResultListener { get; set; }
        /// <summary>
        /// Creates the new dynamic movement.
        /// </summary>
        public void CreateNewDynamicMovement()
        {
            DynamicMovement = new DynamicMovement(Path, ValidationObject);
            DynamicMovement.MovementPrecision = MovementPrecision;
            DynamicMovement.AddProgress(ProgressListener);
        }
        /// <summary>
        /// Creates the new dynamic clash test.
        /// </summary>
        /// <param name="isDynamicRelaxTest">if set to <c>true</c> [is dynamic relax test].</param>
        public void CreateNewDynamicClashTest(bool isDynamicRelaxTest)
        {
            DynamicClashTest = new DynamicClashTest(ValidationObjectItem, BoundaryGeometryItem);
            if (isDynamicRelaxTest)
            {
                DynamicClashTest.ClashTolerance = 50;
                DynamicClashTest.ClashType = ClashTestType.Clearance;
            }
            else
            {
                DynamicClashTest.ClashTolerance = 0.0001;
                DynamicClashTest.ClashType = ClashTestType.Hard;
            }

        }
        /// <summary>
        /// The phase one result collection
        /// </summary>
        public ObservableCollection<Clash3D> PhaseOneResultCollection = new ObservableCollection<Clash3D>();
        /// <summary>
        /// The phase one result real point list
        /// </summary>
        public List<Point3D> PhaseOneResultRealPointList = new List<Point3D>();
        /// <summary>
        /// Runs the first phase of verification. This includes dynamic movement
        /// thorugh path, and clash logging using dynamic clash testing. The result
        /// are saved in the resultcollection as 3D clashes. The real path points after optimalization
        /// are saved in the real point result list.
        /// </summary>
        public bool RunPhaseOneOfVerification()
        {
            CreateNewDynamicMovement();
            CreateNewDynamicClashTest(false);

            //Set init direction
            double angle = 0;
            Rotation3D rot = new Rotation3D(new UnitVector3D(0, 0, 1), angle);

            ValidationObject.SetValidationObjectPosition(Path.CurrentPath.First(), rot);

            //Run phase, and gets results
            PhaseOneResultCollection =  DynamicMovement.MoveValidationObjectThroughPath(DynamicClashTest);

            if (PhaseOneResultCollection == null)
            {
                return false;
            }

            for(int j=0; j < DynamicMovement.RealPath.Count(); j++)
            {
                _previousAnglePath.Add(DynamicMovement.RealPath.ElementAt(j));
            }
            

            PhaseOneResultRealPointList = Path.RunPathOptimalization(
                DynamicMovement.RealPath, 
                PhaseOneResultCollection.ToList());
            return true;
        }
        /// <summary>
        /// The _previous angle path
        /// </summary>
        private List<Point3D> _previousAnglePath = new List<Point3D>();
        /// <summary>
        /// Gets or sets the phase two result collection.
        /// </summary>
        /// <value>The phase two result collection.</value>
        public List<Point3D> PhaseTwoResultCollection { get; set; }

        /// <summary>
        /// Gets or sets the phase two rotation result collection.
        /// </summary>
        /// <value>The phase two rotation result collection.</value>
        public List<Rotation3D> PhaseTwoRotationResultCollection { get; set; }

        private DynamicRelaxationSolver _solver;

        /// <summary>
        /// Runs the second and final phase of verification. This include running
        /// Dynamic relaxation solver with DR algorithm.
        /// </summary>
        /// <param name="newPath">The new path.</param>
        /// <param name="firstTime">if set to <c>true</c> [first time].</param>
        /// <returns><c>true</c> if XXXX, <c>false</c> otherwise.</returns>
        public bool RunPhaseTwoOfVerification(Path newPath, bool firstTime)
        {
            List<Point3D> previousPath = new List<Point3D>();
            if (firstTime)
            {
                previousPath = Path.CurrentPath;
            }
            else
            {
                previousPath = _previousAnglePath;
            }
            

            this.Path = newPath;
            CreateNewDynamicMovement();
            CreateNewDynamicClashTest(true);

            //Getting the last vector, in the previous path. The validation objects current position vector.
            Vector3D preVc = new Vector3D(previousPath.Last().X - previousPath.ElementAt(previousPath.Count() - 2).X,
                previousPath.Last().Y - previousPath.ElementAt(previousPath.Count() - 2).Y,
                previousPath.Last().Z - previousPath.ElementAt(previousPath.Count() - 2).Z);

            //Getting the first vector in the new path. The new position vector of the validation object.
            Vector3D nextVc = new Vector3D(Path.CurrentPath.ElementAt(1).X - Path.CurrentPath.First().X,
                Path.CurrentPath.ElementAt(1).Y - Path.CurrentPath.First().Y,
                Path.CurrentPath.ElementAt(1).Z - Path.CurrentPath.First().Z);

            //Calculate the angle between the two vectors, and rotate the validation object to to correct angle.
            double angle = DynamicMovement.CalculateVectorAngle(preVc, nextVc);

            if (firstTime)
            {
                angle = 0;
            }

            Rotation3D rot = new Rotation3D(new UnitVector3D(0, 0, 1), angle);
            ValidationObject.SetValidationObjectPosition(Path.CurrentPath.First(), rot);

            _solver = new DynamicRelaxationSolver(ValidationObject, Path, MovementPrecision, DynamicClashTest, previousPath);
            _solver.AddProgress(ProgressListener);
            _solver.AddResultListener(ResultListener);
            bool result = _solver.SolveDynamicRelaxation();

            PhaseTwoResultCollection = _solver.RealPath;
            PhaseTwoRotationResultCollection = _solver.RealAnglePath;

            return result;
        }
        /// <summary>
        /// Reverts the start position.
        /// </summary>
        public void RevertStartPosition()
        {
            Path.RevertStartPosition();
        }
        /// <summary>
        /// Cancels the verification.
        /// </summary>
        public void CancelVerification()
        {
            try
            {
                _solver.StopProgress = true;
                Status = Status.Cancelled;
                ChangeStatusIcon();
            }
            catch (Exception e) 
            {
                //do nothing
            }
        }
        /// <summary>
        /// Changes the status icon.
        /// </summary>
        private void ChangeStatusIcon()
        {
            if (Status.Equals(Status.New))
            {
                Icon = "F1 M 38,27.1542C 43.99,27.1542 48.8458,32.01 48.8458,38C 48.8458,43.99 43.99,48.8458 38,48.8458C 32.01,48.8458 27.1542,43.99 27.1542,38C 27.1542,32.01 32.01,27.1542 38,27.1542 Z M 38,16.625C 49.8051,16.625 59.375,26.1949 59.375,38C 59.375,49.8051 49.8051,59.375 38,59.375C 26.1949,59.375 16.625,49.8051 16.625,38C 16.625,26.1949 26.1949,16.625 38,16.625 Z M 38,20.5833C 28.381,20.5833 20.5833,28.381 20.5833,38C 20.5833,47.619 28.381,55.4167 38,55.4167C 47.6189,55.4167 55.4167,47.619 55.4167,38C 55.4167,28.381 47.619,20.5833 38,20.5833 Z ";
                IconColor = "Blue";
            }
            else if (Status.Equals(Status.Working))
            {
                Icon = "F1 M 57,57L 42,57L 42,42L 47.1924,47.1924C 49.5449,44.8399 51,41.5899 51,38C 51,30.8203 45.1797,25 38,25C 30.8203,25 25,30.8203 25,38C 25,42.6137 27.4035,46.6661 31.0266,48.9734L 26.7138,53.2862C 22.0342,49.8254 19,44.2671 19,38C 19,27.5066 27.5066,19 38,19C 48.4934,19 57,27.5066 57,38C 57,43.2467 54.8734,47.9967 51.435,51.435L 57,57 Z ";
                IconColor = "DimGray";
            }
            else if (Status.Equals(Status.Optimized))
            {
                Icon = "F1 M 34.8333,60.1667L 34.8333,57.3958L 41.1667,58.5833L 41.1667,60.1667L 34.8333,60.1667 Z M 31.6666,55.0209L 31.6666,52.25L 44.3333,53.8334L 44.3333,56.6042L 31.6666,55.0209 Z M 44.3333,51.8542L 31.6666,50.2709L 31.6666,47.5L 44.3333,49.0834L 44.3333,51.8542 Z M 38,17.4167C 44.9956,17.4167 50.6667,23.4422 50.6667,30.875C 50.6667,35.8565 44.3333,40.7324 44.3333,42.5329L 44.3333,47.5L 31.6667,45.9167L 31.6667,42.5329C 31.6667,41.1667 25.3333,35.8565 25.3333,30.875C 25.3333,23.4422 31.0044,17.4167 38,17.4167 Z ";
                IconColor = "OliveDrab";
            }
            else if (Status.Equals(Status.Finished))
            {
                Icon = "F1 M 23.7501,33.25L 34.8334,44.3333L 52.2499,22.1668L 56.9999,26.9168L 34.8334,53.8333L 19.0001,38L 23.7501,33.25 Z ";
                IconColor = "LimeGreen";
            }
            else if (Status.Equals(Status.Failed))
            {
                Icon = "F1 M 34.7541,26.4939L 20.5932,1.72809C 19.9132,0.624023 18.9211,0.0480042 17.6171,0C 16.265,0.0480042 15.2729,0.624023 14.6409,1.72809L 0.480042,26.4939C 0.151978,27.0559 -0.00799561,27.6424 0,28.2534C 0.0289917,29.2073 0.378998,29.9982 1.05005,30.6259C 1.72107,31.2536 2.53915,31.579 3.50421,31.6022L 31.7299,31.6022C 32.693,31.5848 33.503,31.271 34.1601,30.6607C 34.8171,30.0504 35.1591,29.248 35.1861,28.2534C 35.1861,27.6424 35.0421,27.0559 34.7541,26.4939 Z M 15.0729,8.06448L 20.2092,8.06448L 20.2092,19.7072L 15.0729,19.7072L 15.0729,8.06448 Z M 17.665,22.4372C 18.4991,22.4576 19.1832,22.7468 19.7172,23.3048C 20.2512,23.8628 20.5272,24.5674 20.5453,25.4186C 20.5272,26.2444 20.2512,26.9266 19.7172,27.4652C 19.1832,28.0039 18.4991,28.2829 17.665,28.3022C 16.831,28.2829 16.147,28.0039 15.6129,27.4653C 15.0789,26.9266 14.8029,26.2444 14.7849,25.4186C 14.8029,24.546 15.0789,23.8353 15.6129,23.2864C 16.147,22.7376 16.831,22.4545 17.665,22.4372 Z ";
                IconColor = "Red";
            }
            else if (Status.Equals(Status.Cancelled))
            {
                Icon = "F1 M 38,19C 48.4934,19 57,27.5066 57,38C 57,48.4934 48.4934,57 38,57C 27.5066,57 19,48.4934 19,38C 19,27.5066 27.5066,19 38,19 Z M 38,23.75C 35.2116,23.75 32.6102,24.5509 30.4134,25.9352L 50.0648,45.5866C 51.4491,43.3898 52.25,40.7884 52.25,38C 52.25,30.13 45.87,23.75 38,23.75 Z M 23.75,38C 23.75,45.8701 30.1299,52.25 38,52.25C 40.7884,52.25 43.3897,51.4491 45.5865,50.0649L 25.9351,30.4136C 24.5509,32.6103 23.75,35.2117 23.75,38 Z ";
                IconColor = "Tomato";
            }
        }
        /// <summary>
        /// The _icon color
        /// </summary>
        private string _iconColor = "White";
        /// <summary>
        /// Gets or sets the color of the icon.
        /// </summary>
        /// <value>The color of the icon.</value>
        public string IconColor
        {
            get { return _iconColor; }
            set
            {
                _iconColor = value;
            }
        }
        /// <summary>
        /// The _icon
        /// </summary>
        private string _icon = "";
        /// <summary>
        /// Gets or sets the icon.
        /// </summary>
        /// <value>The icon.</value>
        public string Icon
        {
            get { return _icon; }
            set
            {
                _icon = value;
            }
        }

        public void RunDynamicMovement()
        {
            List<Point3D> previousPath = PhaseTwoResultCollection;

            //Getting the last vector, in the previous path. The validation objects current position vector.
            Vector3D preVc = new Vector3D(previousPath.Last().X - previousPath.ElementAt(previousPath.Count() - 2).X,
                previousPath.Last().Y - previousPath.ElementAt(previousPath.Count() - 2).Y,
                previousPath.Last().Z - previousPath.ElementAt(previousPath.Count() - 2).Z);

            //Getting the first vector in the new path. The new position vector of the validation object.
            Vector3D nextVc = new Vector3D(PhaseTwoResultCollection.ElementAt(1).X - PhaseTwoResultCollection.First().X,
                PhaseTwoResultCollection.ElementAt(1).Y - PhaseTwoResultCollection.First().Y,
                PhaseTwoResultCollection.ElementAt(1).Z - PhaseTwoResultCollection.First().Z);

            //Calculate the angle between the two vectors, and rotate the validation object to to correct angle.
            double angle = DynamicMovement.CalculateVectorAngle(preVc, nextVc);

            Rotation3D rot = new Rotation3D(new UnitVector3D(0, 0, 1), angle);
            ValidationObject.SetValidationObjectPosition(PhaseTwoResultCollection.First(), rot);

            CreateNewDynamicMovement();
            DynamicMovement.Path.CurrentPath = PhaseTwoResultCollection;
            DynamicMovement.MoveValidationObjectThroughPath(PhaseTwoRotationResultCollection);
        }

        public void CancelOptimization()
        {
            try
            {
                DynamicMovement.StopProgress = true;
                Status = Status.Cancelled;
                ChangeStatusIcon();
            }
            catch (Exception e)
            {
                //do nothing
            }
        }
    }
}
