﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-07-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-17-2014
// ***********************************************************************
// <copyright file="DynamicMovement.cs" >
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Autodesk.Navisworks.Api;
using System.Windows.Threading;
using System.Runtime.InteropServices;
using DynamicRelaxationClashVerification.Models.DataModels;
using System.Collections.ObjectModel;
using System.Threading;

/// <summary>
/// The Models namespace includes all the models and interfaces in the MVVM arcitechture. 
/// </summary>
namespace DynamicRelaxationClashVerification.Models
{
    /// <summary>
    /// The DynamicMovement class has properties and methods for moving the
    /// validation object along a path. It also uses InvalidateRect and UpdateWindow
    /// to update the graphical window in Navisworks for each move.
    /// </summary>
    public class DynamicMovement : ObservableProgress
    {
        /// <summary>
        /// Invalidates the rect.
        /// </summary>
        /// <param name="hWnd">The h WND.</param>
        /// <param name="lpRect">The lp rect.</param>
        /// <param name="bErase">if set to <c>true</c> [b erase].</param>
        /// <returns><c>true</c> if XXXX, <c>false</c> otherwise.</returns>
        [DllImport("user32.dll")]
        static extern bool InvalidateRect(IntPtr hWnd, IntPtr lpRect, bool bErase);
        /// <summary>
        /// Updates the window.
        /// </summary>
        /// <param name="hWnd">The h WND.</param>
        /// <returns><c>true</c> if XXXX, <c>false</c> otherwise.</returns>
        [DllImport("user32.dll")]
        static extern bool UpdateWindow(IntPtr hWnd);
        /// <summary>
        /// The current path to move along.
        /// </summary>
        /// <value>The path.</value>
        public Path Path { get; private set; }
        /// <summary>
        /// The current validation object to move.
        /// </summary>
        /// <value>The validation object.</value>
        public ValidationObject ValidationObject {get; private set;}
        /// <summary>
        /// The movement precision. The step distance for each move.
        /// The smaller number, the more accurate.
        /// </summary>
        /// <value>The movement precision.</value>
        public int MovementPrecision { get; set; }
        /// <summary>
        /// The real path used to store the points after optimalization.
        /// </summary>
        /// <value>The real path.</value>
        public List<Point3D> RealPath { get; set; }
        /// <summary>
        /// Initializes a new instance of the <see cref="DynamicMovement"/> class.
        /// </summary>
        /// <param name="path">The path.</param>
        /// <param name="validationObject">The validation object.</param>
        public DynamicMovement(Path path, ValidationObject validationObject)
        {
            ValidationObject = validationObject;
            Path = path;
            RealPath = new List <Point3D>();
            StopProgress = false;
        }

        public DynamicClashTest ClashTest { get; set; }

        /// <summary>
        /// Moves the validation object through path.
        /// </summary>
        /// <param name="test">The test.</param>
        public ObservableCollection<Clash3D> MoveValidationObjectThroughPath(DynamicClashTest test)
        {
            if (Path.CurrentPath == null || Path.CurrentPath.Count() < 2)
                return null;

            RealPath = new List<Point3D>();

            ProgressMin = 0;
            ProgressMax = Path.CurrentPath.Count();
            FireProgressChanged(0);

            int testCounter = 1;

            Vector3D previousVc = 
                new Vector3D(
                    Path.CurrentPath.ElementAt(1).ToVector3D().X - Path.CurrentPath.First().ToVector3D().X, 
                    Path.CurrentPath.ElementAt(1).ToVector3D().Y - Path.CurrentPath.First().ToVector3D().Y,
                    0);

            RealPath.Add(
                new Point3D(
                    Path.CurrentPath.First().X,
                    Path.CurrentPath.First().Y,
                    0));

            for (int i = 1; i < Path.CurrentPath.Count(); i++)
            {
                double x1 = Path.CurrentPath.ElementAt(i - 1).ToVector3D().X;
                double y1 = Path.CurrentPath.ElementAt(i - 1).ToVector3D().Y;

                double x2 = Path.CurrentPath.ElementAt(i).ToVector3D().X;
                double y2 = Path.CurrentPath.ElementAt(i).ToVector3D().Y;

                //Vector from the current point to next point in Geometry.
                Vector3D vcL = new Vector3D(x2 - x1, y2 - y1, 0);
                
                //Creating unit vector from the current vector.
                UnitVector3D unitVc = new UnitVector3D(vcL);

                //Scaling the unitvector for faster calculations
                Vector3D scaledUnitVc = 
                    new Vector3D(
                        unitVc.X * MovementPrecision,
                        unitVc.Y * MovementPrecision,
                        0);

                //Rotate the validation object based on the direction of the vector
                double angle = CalculateVectorAngle(previousVc, vcL);
                Rotation3D rot = new Rotation3D(new UnitVector3D(0, 0, 1), angle);

                ValidationObject.RotateValidationObject(rot, RealPath.Last());
                
                double X_cur = RealPath.Last().X;
                double Y_cur = RealPath.Last().Y;

                Vector3D dvL = new Vector3D(x2 - X_cur, y2 - Y_cur, 0);

                //While the validation object havn't reached the next point.
                while (vcL.Dot(dvL) >= 0)
                {

                    if (StopProgress == true)
                    {
                        return null;
                    }

                    //Code for creating the "animation effect".
                    DoEvents();

                    //Makes the UI update everytime MoveValidationObject is called
                    IntPtr hWnd = Autodesk.Navisworks.Api.Application.Gui.MainWindow.Handle;
                    InvalidateRect(hWnd, IntPtr.Zero, true);
                    UpdateWindow(hWnd);

                    //Move the validation object, based on the scaled unit vector. 
                    ValidationObject.MoveValidationObject(scaledUnitVc);
                    if (test == null)
                    {
                        Thread.Sleep(100);
                    }

                    //Adds the current point.
                    RealPath.Add(
                        new Point3D(
                            RealPath.Last().X + scaledUnitVc.X, 
                            RealPath.Last().Y + scaledUnitVc.Y, 
                            0));

                    X_cur = RealPath.Last().X;
                    Y_cur = RealPath.Last().Y;
                    dvL = new Vector3D(x2 - X_cur, y2 - Y_cur, 0);

                    //If we should run clash test (not just move validation object)
                    if (test != null)
                    {
                        //Creates and runs clash test
                        test.CreateDynamicClashTest();
                        test.RunClashTest();

                        //If any clashes accured. Registrat the current point of the validation object during the clash.
                        //The clashtest gives only the clash point.
                        if (test.IsClashes)
                        {
                            test.ClashCollection.Last().RealPathPoint = RealPath.Last();
                            test.ClashCollection.Last().RealPathPointIndex = RealPath.IndexOf(RealPath.Last());
                        }
                        testCounter++;
                    }
                }

                if (x2 - X_cur > 0 || y2 - Y_cur > 0)
                {
                    //Code for creating the "animation effect".
                    DoEvents();

                    //Makes the UI update everytime MoveValidationObject is called
                    IntPtr hWnd = Autodesk.Navisworks.Api.Application.Gui.MainWindow.Handle;
                    InvalidateRect(hWnd, IntPtr.Zero, true);
                    UpdateWindow(hWnd);

                    Vector3D shortVc = new Vector3D(x2 - X_cur, y2 - Y_cur, 0);

                    //Move the validation object, based on the scaled unit vector. 
                    ValidationObject.MoveValidationObject(shortVc);
                    if (test == null)
                    {
                        Thread.Sleep(100);
                    }

                    //Adds the current point.
                    RealPath.Add(
                        new Point3D(
                            RealPath.Last().X + shortVc.X,
                            RealPath.Last().Y + shortVc.Y,
                            0));

                    X_cur = RealPath.Last().X;
                    Y_cur = RealPath.Last().Y;

                    //If we should run clash test (not just move validation object)
                    if (test != null)
                    {
                        //Creates and runs clash test
                        test.CreateDynamicClashTest();
                        test.RunClashTest();

                        //If any clashes accured. Registrat the current point of the validation object during the clash.
                        //The clashtest gives only the clash point.
                        if (test.IsClashes)
                        {
                            test.ClashCollection.Last().RealPathPoint = RealPath.Last();
                            test.ClashCollection.Last().RealPathPointIndex = RealPath.IndexOf(RealPath.Last());
                        }
                        testCounter++;
                    }
                }
                previousVc = vcL;
                FireProgressChanged(i);
            }
            FireProgressChanged(Path.CurrentPath.Count());

            if (test == null)
            {
                return null;
            }
            else
            {
                return test.ClashCollectionList;
            }
        }

        public void MoveValidationObjectThroughPath(List<Rotation3D> RotationPath)
        {

            for(int i=1; i< Path.CurrentPath.Count();i++)
            {
                Point3D point_1 = Path.CurrentPath.ElementAt(i-1);
                Point3D point_2 = Path.CurrentPath.ElementAt(i);

                double x1 = point_1.ToVector3D().X;
                double y1 = point_1.ToVector3D().Y;

                double x2 = point_2.ToVector3D().X;
                double y2 = point_2.ToVector3D().Y;

                //Vector from the current point to next point in Geometry.
                Vector3D vc = new Vector3D(x2 - x1, y2 - y1, 0);

                //Creating unit vector from the current vector.
                UnitVector3D unitVc = new UnitVector3D(vc);

                //Scaling the unitvector for faster calculations
                Vector3D scaledUnitVc =
                    new Vector3D(
                        unitVc.X * MovementPrecision,
                        unitVc.Y * MovementPrecision,
                        0);

                //Code for creating the "animation effect".
                DoEvents();

                //Makes the UI update everytime MoveValidationObject is called
                IntPtr hWnd = Autodesk.Navisworks.Api.Application.Gui.MainWindow.Handle;
                InvalidateRect(hWnd, IntPtr.Zero, true);
                UpdateWindow(hWnd);

                ValidationObject.RotateValidationObject(RotationPath.ElementAt(i-1), Path.CurrentPath.ElementAt(i - 1));

                //Code for creating the "animation effect".
                DoEvents();

                //Makes the UI update everytime MoveValidationObject is called
                InvalidateRect(hWnd, IntPtr.Zero, true);
                UpdateWindow(hWnd);

                //Move the validation object, based on the scaled unit vector. 
                ValidationObject.MoveValidationObject(vc);
                Thread.Sleep(500);
            }
        }

        /// <summary>
        /// Calculates the vector angle.
        /// </summary>
        /// <param name="previousVc">The previous vc.</param>
        /// <param name="vc">The vc.</param>
        /// <returns>System.Double.</returns>
        public static double CalculateVectorAngle(Vector3D previousVc, Vector3D vc)
        {
            UnitVector3D u = new UnitVector3D(previousVc);
            UnitVector3D v = new UnitVector3D(vc);

            return u.AngleWithReference(v, new UnitVector3D(0, 0, 1));
        }

        public bool StopProgress { get; set; }
    }
}
