﻿// ***********************************************************************
// Assembly         : DynamicClash.MHRK
// Author           : mariuroe
// Created          : 03-31-2014
//
// Last Modified By : mariuroe
// Last Modified On : 04-16-2014
// ***********************************************************************
// <copyright file="DynamicRelaxationSolver.cs">
//     Copyright (c) . All rights reserved.
// </copyright>
// <summary></summary>
// ***********************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Autodesk.Navisworks.Api;
using System.Runtime.InteropServices;
using System.Windows.Threading;
using DynamicRelaxationClashVerification.Models.DataModels;

/// <summary>
/// The Models namespace includes all the models and interfaces in the MVVM arcitechture. 
/// </summary>
namespace DynamicRelaxationClashVerification.Models
{
    /// <summary>
    /// The DynamicRelaxationSolver class contains the dynamic relaxation algorithme for
    /// solving the verification problem. The class inherits the abstract class ObservableProgress,
    /// for monitor the progress over the algorithm. More information about the matemathic and
    /// physics used in the algorithm, and how it works, see the master thesis report.
    /// </summary>
    public class DynamicRelaxationSolver : ObservableProgress
    {
        /// <summary>
        /// Invalidates the rect.
        /// </summary>
        /// <param name="hWnd">The h WND.</param>
        /// <param name="lpRect">The lp rect.</param>
        /// <param name="bErase">if set to <c>true</c> [b erase].</param>
        /// <returns><c>true</c> if XXXX, <c>false</c> otherwise.</returns>
        [DllImport("user32.dll")]
        static extern bool InvalidateRect(IntPtr hWnd, IntPtr lpRect, bool bErase);
        /// <summary>
        /// Updates the window.
        /// </summary>
        /// <param name="hWnd">The h WND.</param>
        /// <returns><c>true</c> if XXXX, <c>false</c> otherwise.</returns>
        [DllImport("user32.dll")]
        static extern bool UpdateWindow(IntPtr hWnd);
        /// <summary>
        /// The current path, used for getting drag points in DR.
        /// </summary>
        /// <value>The path.</value>
        public Path Path { get; private set; }
        /// <summary>
        /// The current validation object.
        /// </summary>
        /// <value>The validation object.</value>
        public ValidationObject ValidationObject { get; private set; }
        /// <summary>
        /// The movement precision.
        /// </summary>
        /// <value>The movement precision.</value>
        public int MovementPrecision { get; set; }
        /// <summary>
        /// The real path to store the final calculated points.
        /// </summary>
        /// <value>The real path.</value>
        public List<Point3D> RealPath { get; set; }
        /// <summary>
        /// Gets or sets the real angle path.
        /// </summary>
        /// <value>The real angle path.</value>
        public List<Rotation3D> RealAnglePath { get; set; }
        /// <summary>
        /// The stiffness of the boundary. It is used to increase the stiffness
        /// of the validation object, when clashing. It is set to: 8 (should not be changed, if
        /// you doesn't know what you are doing).
        /// </summary>
        /// <value>The stiffness boundary.</value>
        private double StiffnessBoundary { get; set; }
        /// <summary>
        /// The damping for the boundary. It is used to increase the damping of the validation 
        /// object when clashing. It is set to: 2 * Math.Sqr(Mass * Sum(All Stiffnesses)) * 2
        /// </summary>
        /// <value>The boundary damping.</value>
        private double BoundaryDamping { get; set; }
        /// <summary>
        /// Gets or sets the dynamic clash test.
        /// </summary>
        /// <value>The dynamic clash test.</value>
        DynamicClashTest DynamicClashTest { get; set; }
        /// <summary>
        /// Initializes a new instance of the <see cref="DynamicRelaxationSolver"/> class.
        /// </summary>
        /// <param name="validationObject">The validation object.</param>
        /// <param name="path">The path.</param>
        /// <param name="movementPrecision">The movement precision.</param>
        /// <param name="dynamicClashTest">The dynamic clash test.</param>
        public DynamicRelaxationSolver(
                                        ValidationObject validationObject, 
                                        Path path, int movementPrecision, 
                                        DynamicClashTest dynamicClashTest,
                                        List<Point3D> previousPath
            )
        {

            if (previousPath != null)
            {
                PreviousPath = previousPath;
            }
            else
            {
                PreviousPath = path.CurrentPath; 
            }

            ValidationObject = validationObject;
            Path = path;
            MovementPrecision = movementPrecision;
            DynamicClashTest = dynamicClashTest;
            RealPath = new List<Point3D>();
            RealAnglePath = new List<Rotation3D>();
            StopProgress = false;
        }
        /// <summary>
        /// The current node
        /// </summary>
        Point3D _currentNode;
        /// <summary>
        /// The drag node
        /// </summary>
        Point3D _dragNode;
        /// <summary>
        /// Initializes the validation object properties, incl. mass, stiffness, damping.
        /// </summary>
        private void InitValidationObjectProperties()
        {
            ValidationObject.Mass = 2.2;
            ValidationObject.Stiffness = 1.1;
            ValidationObject.Damping = 2.2 * Math.Sqrt(ValidationObject.Mass * ValidationObject.Stiffness) * 2;
        }
        /// <summary>
        /// Initializes the boundary geomerty properties.
        /// </summary>
        private void InitBoundaryGeomertyProperties()
        {
            StiffnessBoundary = 8;
            BoundaryDamping =  Math.Abs(ValidationObject.Mass * (ValidationObject.Stiffness + StiffnessBoundary))  - ValidationObject.Damping;
        }

        private List<Point3D> PreviousPath { get; set; }

        private double _angle = 0;
        /// <summary>
        /// Solves the dynamic relaxation.
        /// </summary>
        /// <returns><c>true</c> if verification was possible, <c>false</c> otherwise.</returns>
        public bool SolveDynamicRelaxation()
        {
            if (Path.CurrentPath == null || Path.CurrentPath.Count() < 2)
                return false;

            InitValidationObjectProperties();
            InitBoundaryGeomertyProperties();

            ProgressMin = 0;
            ProgressMax = Path.CurrentPath.Count();

            _currentNode = 
            new Point3D(
                    Path.CurrentPath.First().X,
                    Path.CurrentPath.First().Y,
                    0);

            RealPath.Add(_currentNode);
            Rotation3D rote = new Rotation3D(new UnitVector3D(0, 0, 1), 0);
            RealAnglePath.Add(rote);

            Vector3D currentVc =
                new Vector3D(
                    Path.CurrentPath.ElementAt(1).ToVector3D().X - _currentNode.X,
                    Path.CurrentPath.ElementAt(1).ToVector3D().Y - _currentNode.Y,
                    0);

            Point3D previousPoint = 
                new Point3D(
                    PreviousPath.First().X,
                    PreviousPath.First().Y,
                    0);

            for (int i = 1; i < Path.CurrentPath.Count(); i++)
            {
                _dragNode = Path.CurrentPath.ElementAt(i);

                if (_currentNode.DistanceTo(_dragNode) < MovementPrecision)
                {
                    continue;
                }

                double x1 = previousPoint.ToVector3D().X;
                double y1 = previousPoint.ToVector3D().Y;

                previousPoint = PreviousPath.ElementAt(i);

                double x2 = PreviousPath.ElementAt(i).X;
                double y2 = PreviousPath.ElementAt(i).Y;

                Vector3D vc = new Vector3D(x2 - x1, y2 - y1, 0);

                //Rotate the validation object based on the direction of the vector
                _angle = DynamicMovement.CalculateVectorAngle(currentVc, vc);
                Rotation3D rot = new Rotation3D(new UnitVector3D(0, 0, 1), _angle);
                ValidationObject.RotateValidationObject(rot, RealPath.Last());

                if (NextIteration() && StopProgress != true)
                {
                    currentVc = vc;
                    FireProgressChanged(i);
                    continue;
                }
                else
                {
                    FireProgressChanged(Path.CurrentPath.Count());
                    return false;
                }
            }
            FireProgressChanged(Path.CurrentPath.Count());
            return true;
        }
        /// <summary>
        /// The h WND
        /// </summary>
        private IntPtr hWnd = Autodesk.Navisworks.Api.Application.Gui.MainWindow.Handle;
        /// <summary>
        /// Next iteration in DR.
        /// </summary>
        /// <returns><c>true</c> if can move to drag node, <c>false</c> otherwise.</returns>
        private bool NextIteration()
        {
            ValidationObject.InitAllDynamicProperties();

            bool firstIteration = true;
            int counter = 0;

            double resultantX = 0, resultantY = 0;
            while (!(Math.Abs(resultantX) < 1e-1 && Math.Abs(resultantY) < 1e-1) || firstIteration)
            {
                firstIteration = false;
                counter++;
                ValidationObject.InitForcesAndMoments();

                //Calculating displacement between real point and drag point
                Vector3D displacement = new Vector3D(
                    _currentNode.X - _dragNode.X,
                    _currentNode.Y - _dragNode.Y,
                    _currentNode.Z - _dragNode.Z);

                //Creates and runs clash test
                DynamicClashTest.CreateDynamicClashTest();
                DynamicClashTest.RunClashTest();

                Vector3D vc = new Vector3D(0,0,0);
                UnitVector3D uVc = new UnitVector3D(0,0,0);
                double clashDistance = 0.0;
                double deltaT = 0.7;
                double damp = 0;

                //Calculating forces based on clashes
                //from Navisworks clash API.

                foreach (Clash3D clash in DynamicClashTest.ClashCollection)
                {
                    Point3D clashPoint = clash.ClashPoint;
                    Point3D pathPoint = _dragNode;

                    if (Math.Sign(clash.Distance) > 0)
                    {
                        clashDistance = 50 - Math.Abs(clash.Distance);
                    }
                    else
                    {
                        clashDistance = 50 + Math.Abs(clash.Distance);
                    }
                        
                        deltaT = 0.09;
                        damp = BoundaryDamping;

                    vc = clash.GetClashDirectionVector();

                    uVc = new UnitVector3D(vc);

                    double normVel = uVc.Dot(new Vector3D(ValidationObject.Velocity[0], ValidationObject.Velocity[1],0));

                    ValidationObject.Force[0] += StiffnessBoundary * uVc.X * clashDistance - damp * normVel * uVc.X;
                    ValidationObject.Force[1] += StiffnessBoundary * uVc.Y * clashDistance - damp * normVel * uVc.Y; 
                }

                double alpha = 0.1;

                double dampingLeddX = ValidationObject.Damping * ValidationObject.Velocity[0];
                double stiffnessLeddX = ValidationObject.Stiffness * displacement.X;

                resultantX = alpha * resultantX + (1.0-alpha)*(ValidationObject.Force[0] - 
                    stiffnessLeddX - 
                    dampingLeddX);

                ValidationObject.Acceleration[0] = (1 / ValidationObject.Mass) * resultantX;
                ValidationObject.Velocity[0] = ValidationObject.Velocity[0] +
                    ValidationObject.Acceleration[0] * deltaT;

                double nextXpos = _currentNode.X + ValidationObject.Velocity[0] * deltaT;

                double dampingLeddY = ValidationObject.Damping * ValidationObject.Velocity[1];
                double stiffnessLeddY = ValidationObject.Stiffness * displacement.Y;

                resultantY = alpha * resultantY + (1.0-alpha) *(ValidationObject.Force[1] -
                    stiffnessLeddY -
                    dampingLeddY);

                ValidationObject.Acceleration[1] = (1 / ValidationObject.Mass) * resultantY;
                ValidationObject.Velocity[1] = ValidationObject.Velocity[1] +
                    ValidationObject.Acceleration[1] * deltaT;

                double nextYpos = _currentNode.Y + ValidationObject.Velocity[1] * deltaT;

                Point3D newNode = new Point3D(nextXpos, nextYpos, 0);

                FireResultChanged(
                    new Point2D(ValidationObject.Velocity[0], ValidationObject.Velocity[1]),
                    new Point2D(ValidationObject.Acceleration[0], ValidationObject.Acceleration[1]),
                    new Point2D(ValidationObject.Force[0], ValidationObject.Force[1]));

                //Try to move validation object

                //Code for creating the "animation effect".
                DoEvents();

                MoveValidationObjectBasedOnTwoPoints(_currentNode, newNode);

                //Makes the UI update everytime MoveValidationObject is called
                InvalidateRect(hWnd, IntPtr.Zero, true);
                UpdateWindow(hWnd);

                _currentNode = newNode;

                if (counter > 500 || StopProgress==true)
                {
                    RealPath.Add(_currentNode);
                    return false;
                }
            }

            ////Makes the UI update everytime MoveValidationObject is called
            InvalidateRect(hWnd, IntPtr.Zero, true);
            UpdateWindow(hWnd);

            RealPath.Add(_currentNode);

            Rotation3D rot = new Rotation3D(new UnitVector3D(0, 0, 1), _angle);
            RealAnglePath.Add(rot);

            return true;
        }
        /// <summary>
        /// Help method: Moves the validation object from point_1 to point_2.
        /// </summary>
        /// <param name="point_1">The point_1.</param>
        /// <param name="point_2">The point_2.</param>
        private  void MoveValidationObjectBasedOnTwoPoints(Point3D point_1, Point3D point_2)
        {
            double x1 = point_1.ToVector3D().X;
            double y1 = point_1.ToVector3D().Y;

            double x2 = point_2.ToVector3D().X;
            double y2 = point_2.ToVector3D().Y;

            //Vector from the current point to next point in Geometry.
            Vector3D vc = new Vector3D(x2 - x1, y2 - y1, 0);

            //Creating unit vector from the current vector.
            UnitVector3D unitVc = new UnitVector3D(vc);

            //Scaling the unitvector for faster calculations
            Vector3D scaledUnitVc =
                new Vector3D(
                    unitVc.X * MovementPrecision,
                    unitVc.Y * MovementPrecision,
                    0);

            //Move the validation object, based on the scaled unit vector. 
            ValidationObject.MoveValidationObject(vc);

        }

        /// <summary>
        /// Gets or sets a value indicating whether [stop progress].
        /// </summary>
        /// <value><c>true</c> if [stop progress]; otherwise, <c>false</c>.</value>
        public bool StopProgress { get; set; }
    }
}
